import 'dart:async';

import 'package:equatable/equatable.dart';
import 'package:osam/osam.dart';
import 'package:osam/src/util/value_stream.dart';

typedef ValueMapper<ST, V> = V Function(ST state);
typedef StateMapper<ST> = void Function(ST state);

// ignore: must_be_immutable
abstract class BaseState<ST> extends Equatable {
  StreamController<ST> _stateBroadcaster;
  int _lastKnownHashCode = -1;

  void _initController() {
    _stateBroadcaster = StreamController<ST>.broadcast(onCancel: () {
      _stateBroadcaster.close();
    });
  }

  ValueStream<ST> get _stateStream => ValueStream(_stateBroadcaster.stream.map((state) => state), this as ST);

  ValueStream<V> propertyStream<V>(ValueMapper<ST, V> mapper) {
    if (_stateBroadcaster == null || _stateBroadcaster.isClosed) _initController();
    return ValueStream(
        (PropertyStream<V>(
                _stateStream.stream.map<V>((state) {
                  try {
                    return mapper(state);
                  } catch (e) {
                    handleError(e);
                    return null;
                  }
                }).where((property) => property != null),
                mapper(_stateStream.value)))
            .propertyStream,
        mapper(_stateStream.value));
  }

  void update(StateMapper<ST> change) {
    if (_stateBroadcaster == null || _stateBroadcaster.isClosed) _initController();
    try {
      change(this as ST);
    } catch (e) {
      handleError(e);
    }
    if (_lastKnownHashCode != hashCode) {
      _stateBroadcaster.sink.add(this as ST);
      _lastKnownHashCode = hashCode;
    }
  }

  void handleError(Object error) => print(error);

  void throwError(Object error) async => await _stateBroadcaster.addError(error);

  void clear() {}

  @override
  List<Object> get props;
}

class PropertyStream<V> {
  final Stream<V> _inputStream;
  final V _firstValue;
  int _valueHashCode = -1;

  PropertyStream(this._inputStream, this._firstValue) {
    _valueHashCode = _ComparableWrapper(_firstValue).hashCode;
  }

  Stream<V> get propertyStream => _inputStream.map((next) {
        final comparableWrapper = _ComparableWrapper(next);
        final result = _valueHashCode == comparableWrapper.hashCode;
        _valueHashCode = comparableWrapper.hashCode;
        return result ? null : next;
      }).where((event) => event != null);
}

class _ComparableWrapper<V> extends Equatable {
  final V _value;

  _ComparableWrapper(this._value);

  @override
  List<Object> get props => [_value];
}
